/**
  ******************************************************************************
  * @file    stm32f7_som.c
  * @author  Vladimir Skvortsov <vskvortsov@emcraft.com>
  * @version V1.0.0
  * @date    29-December-2017
  * @brief   This file includes a standard driver for the M12L256 SDRAM
  *          memory device mounted on STM32F7-SOM boards.
  ******************************************************************************
  * @attention
  *
  * <h2><center>&copy; COPYRIGHT(c) 2017 EmCraft Systems</center></h2>
  * <h2><center>&copy; COPYRIGHT(c) 2016 STMicroelectronics</center></h2>
  *
  * Redistribution and use in source and binary forms, with or without modification,
  * are permitted provided that the following conditions are met:
  *   1. Redistributions of source code must retain the above copyright notice,
  *      this list of conditions and the following disclaimer.
  *   2. Redistributions in binary form must reproduce the above copyright notice,
  *      this list of conditions and the following disclaimer in the documentation
  *      and/or other materials provided with the distribution.
  *   3. Neither the name of STMicroelectronics nor the names of its contributors
  *      may be used to endorse or promote products derived from this software
  *      without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
  ******************************************************************************
  */

#include "stm32f7_som_sdram.h"
#include "stm32f7xx_hal_sdram.h"

/* Forward declarations */
static void BSP_SDRAM_Initialization_sequence(SDRAM_HandleTypeDef * hsdram,
					      uint32_t RefreshCount);
static void BSP_SDRAM_GPIOInit(void);

#define SDRAM_TIMEOUT                    ((uint32_t)0xFFFF)
#define max(X, Y) ( (X) > (Y) ? (X) : (Y) )

/**
  * @brief  FMC SDRAM Mode definition register defines
  */
#define SDRAM_MODEREG_BURST_LENGTH_1             ((uint16_t)0x0000)
#define SDRAM_MODEREG_BURST_LENGTH_2             ((uint16_t)0x0001)
#define SDRAM_MODEREG_BURST_LENGTH_4             ((uint16_t)0x0002)
#define SDRAM_MODEREG_BURST_LENGTH_8             ((uint16_t)0x0004)
#define SDRAM_MODEREG_BURST_TYPE_SEQUENTIAL      ((uint16_t)0x0000)
#define SDRAM_MODEREG_BURST_TYPE_INTERLEAVED     ((uint16_t)0x0008)
#define SDRAM_MODEREG_CAS_LATENCY_2              ((uint16_t)0x0020)
#define SDRAM_MODEREG_CAS_LATENCY_3              ((uint16_t)0x0030)
#define SDRAM_MODEREG_OPERATING_MODE_STANDARD    ((uint16_t)0x0000)
#define SDRAM_MODEREG_WRITEBURST_MODE_PROGRAMMED ((uint16_t)0x0000)
#define SDRAM_MODEREG_WRITEBURST_MODE_SINGLE     ((uint16_t)0x0200)

/*
 * Following are timings for M12L2561616A-6BI, from corresponding datasheet
 */
static uint32_t _ns2clk(uint32_t ns, uint32_t freq)
{
	uint32_t tmp = freq / 1000000;
	return (tmp * ns) / 1000;
}

#define NS2CLK(ns) (_ns2clk(ns, freq))

#define SDRAM_TRRD	NS2CLK(12)
#define SDRAM_TRCD	NS2CLK(18)
#define SDRAM_TRP	NS2CLK(18)
#define SDRAM_TRAS	NS2CLK(42)
#define SDRAM_TRC	NS2CLK(60)
#define SDRAM_TRFC	NS2CLK(60)
#define SDRAM_TCDL	(1 - 1)
#define SDRAM_TRDL	NS2CLK(12)
#define SDRAM_TBDL	(1 - 1)
#define SDRAM_TREF	(NS2CLK(64000000 / 8192) - 20)
#define SDRAM_TCCD	(1 - 1)

#define SDRAM_TXSR	SDRAM_TRFC	/* Row cycle time after precharge */
#define SDRAM_TMRD	(3 - 1)	/* Page 10, Mode Register Set */

/* Last data in to row precharge, need also comply ineq on page 1648 */
#define SDRAM_TWR	(max(						\
	(int)max((int)SDRAM_TRDL, (int)(SDRAM_TRAS - SDRAM_TRCD)), \
	(int)(SDRAM_TRC - SDRAM_TRCD - SDRAM_TRP)	\
	))

/** @defgroup STM32F769I_DISCOVERY_SDRAM_Private_Variables SDRAM Private Variables
  * @{
  */
static SDRAM_HandleTypeDef sdramHandle;

/**
  * @brief  Initializes the SDRAM device.
  * @retval SDRAM status
  */
uint8_t BSP_SDRAM_Init(void)
{
	uint32_t freq;
	SDRAM_HandleTypeDef hdsram;
	FMC_SDRAM_TimingTypeDef timing;

	freq = HAL_RCC_GetHCLKFreq() / 2;

	sdramHandle.Init.SDBank = FMC_SDRAM_BANK1;
	sdramHandle.Init.ColumnBitsNumber = FMC_SDRAM_COLUMN_BITS_NUM_9;
	sdramHandle.Init.RowBitsNumber = FMC_SDRAM_ROW_BITS_NUM_13;
	sdramHandle.Init.MemoryDataWidth = FMC_SDRAM_MEM_BUS_WIDTH_16;
	sdramHandle.Init.InternalBankNumber = FMC_SDRAM_INTERN_BANKS_NUM_4;
	sdramHandle.Init.CASLatency = FMC_SDRAM_CAS_LATENCY_3;
	sdramHandle.Init.WriteProtection = FMC_SDRAM_WRITE_PROTECTION_DISABLE;
	sdramHandle.Init.SDClockPeriod = FMC_SDRAM_CLOCK_PERIOD_2;
	sdramHandle.Init.ReadBurst = FMC_SDRAM_RBURST_ENABLE;
	sdramHandle.Init.ReadPipeDelay = FMC_SDRAM_RPIPE_DELAY_0;

	sdramHandle.Instance = FMC_SDRAM_DEVICE;

	timing.LoadToActiveDelay = SDRAM_TMRD + 1;
	timing.ExitSelfRefreshDelay = SDRAM_TXSR + 1;
	timing.SelfRefreshTime = SDRAM_TRAS + 1;
	timing.RowCycleDelay = SDRAM_TRC + 1;
	timing.WriteRecoveryTime = SDRAM_TWR + 1;
	timing.RPDelay = SDRAM_TRP + 1;
	timing.RCDDelay = SDRAM_TRCD + 1;

	__HAL_RCC_FMC_CLK_ENABLE();

	/* SDRAM controller initialization */
	BSP_SDRAM_GPIOInit();

	if (HAL_SDRAM_Init(&sdramHandle, &timing) != HAL_OK) {
		return SDRAM_ERROR;
	}

	BSP_SDRAM_Initialization_sequence(&hdsram, SDRAM_TREF / 2);

	return SDRAM_OK;
}

/**
  * @brief  DeInitializes the SDRAM device.
  * @retval SDRAM status
  */
uint8_t BSP_SDRAM_DeInit(void)
{
	/* SDRAM device de-initialization */
	sdramHandle.Instance = FMC_SDRAM_DEVICE;

	if (HAL_SDRAM_DeInit(&sdramHandle) != HAL_OK) {
		return ERROR;
	}

	return SDRAM_OK;
}

/**
  * @brief  Reads an amount of data from the SDRAM memory in polling mode.
  * @param  uwStartAddress: Read start address
  * @param  pData: Pointer to data to be read
  * @param  uwDataSize: Size of read data from the memory
  * @retval SDRAM status
  */
uint8_t BSP_SDRAM_ReadData(uint32_t uwStartAddress, uint32_t * pData,
			   uint32_t uwDataSize)
{
	if (HAL_SDRAM_Read_32b
	    (&sdramHandle, (uint32_t *) uwStartAddress, pData,
	     uwDataSize) != HAL_OK) {
		return SDRAM_ERROR;
	} else {
		return SDRAM_OK;
	}
}

/**
  * @brief  Writes an amount of data to the SDRAM memory in polling mode.
  * @param  uwStartAddress: Write start address
  * @param  pData: Pointer to data to be written
  * @param  uwDataSize: Size of written data from the memory
  * @retval SDRAM status
  */
uint8_t BSP_SDRAM_WriteData(uint32_t uwStartAddress, uint32_t * pData,
			    uint32_t uwDataSize)
{
	if (HAL_SDRAM_Write_32b
	    (&sdramHandle, (uint32_t *) uwStartAddress, pData,
	     uwDataSize) != HAL_OK) {
		return SDRAM_ERROR;
	} else {
		return SDRAM_OK;
	}
}

/**
  * @brief  Initializes FMC pins.
  * @param  None
  * @retval None
  */
static void BSP_SDRAM_GPIOInit(void)
{
	GPIO_InitTypeDef gpio_init_structure;

	/* Enable GPIOs clock */
	__HAL_RCC_GPIOD_CLK_ENABLE();
	__HAL_RCC_GPIOE_CLK_ENABLE();
	__HAL_RCC_GPIOF_CLK_ENABLE();
	__HAL_RCC_GPIOG_CLK_ENABLE();
	__HAL_RCC_GPIOH_CLK_ENABLE();
	__HAL_RCC_GPIOC_CLK_ENABLE();

	/* Common GPIO configuration */
	gpio_init_structure.Mode = GPIO_MODE_AF_PP;
	gpio_init_structure.Pull = GPIO_NOPULL;
	gpio_init_structure.Speed = GPIO_SPEED_HIGH;
	gpio_init_structure.Alternate = GPIO_AF12_FMC;

	/* GPIOD configuration */
	gpio_init_structure.Pin =
	    GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_4 | GPIO_PIN_5 | GPIO_PIN_8 |
	    GPIO_PIN_9 | GPIO_PIN_10 | GPIO_PIN_11 | GPIO_PIN_12 | GPIO_PIN_13 |
	    GPIO_PIN_14 | GPIO_PIN_15;

	HAL_GPIO_Init(GPIOD, &gpio_init_structure);

	/* GPIOE configuration */
	gpio_init_structure.Pin =
	    GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_3 | GPIO_PIN_4 | GPIO_PIN_5 |
	    GPIO_PIN_6 | GPIO_PIN_7 | GPIO_PIN_8 | GPIO_PIN_9 | GPIO_PIN_10 |
	    GPIO_PIN_11 | GPIO_PIN_12 | GPIO_PIN_13 | GPIO_PIN_14 | GPIO_PIN_15;

	HAL_GPIO_Init(GPIOE, &gpio_init_structure);

	/* GPIOF configuration */
	gpio_init_structure.Pin =
	    GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_2 | GPIO_PIN_3 | GPIO_PIN_4 |
	    GPIO_PIN_5 | GPIO_PIN_11 | GPIO_PIN_12 | GPIO_PIN_13 | GPIO_PIN_14 |
	    GPIO_PIN_15;

	HAL_GPIO_Init(GPIOF, &gpio_init_structure);

	/* GPIOG configuration */
	gpio_init_structure.Pin =
	    GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_2 | GPIO_PIN_4 | GPIO_PIN_5 |
	    GPIO_PIN_8 | GPIO_PIN_15;

	HAL_GPIO_Init(GPIOG, &gpio_init_structure);

	/* GPIOH configuration */
	gpio_init_structure.Pin = GPIO_PIN_5;

	HAL_GPIO_Init(GPIOH, &gpio_init_structure);

	/* GPIOC configuration */
	gpio_init_structure.Pin = GPIO_PIN_2 | GPIO_PIN_3;

	HAL_GPIO_Init(GPIOC, &gpio_init_structure);
}

/**
  * @brief  Programs the SDRAM device.
  * @param  RefreshCount: SDRAM refresh counter value
  * @retval None
  */
static void BSP_SDRAM_Initialization_sequence(SDRAM_HandleTypeDef * hsdram,
					      uint32_t RefreshCount)
{
	static FMC_SDRAM_CommandTypeDef Command;
	int i;

	/* Step 1: Configure a clock configuration enable command */
	Command.CommandMode = FMC_SDRAM_CMD_CLK_ENABLE;
	Command.CommandTarget = FMC_SDRAM_CMD_TARGET_BANK1;
	Command.AutoRefreshNumber = 1;
	Command.ModeRegisterDefinition = 0;

	/* Send the command */
	HAL_SDRAM_SendCommand(&sdramHandle, &Command, SDRAM_TIMEOUT);

	/* Step 2: Insert 100 us minimum delay */
	/* Working with 200MHz freq we exec max 200 instructions per 1 us, so
	 * the following iteration cycle will definitely take more that 100 us
	 */
	for (i = 0; i < 20000; i++) ;

	/* Step 3: Configure a PALL (precharge all) command */
	Command.CommandMode = FMC_SDRAM_CMD_PALL;
	Command.CommandTarget = FMC_SDRAM_CMD_TARGET_BANK1;
	Command.AutoRefreshNumber = 1;
	Command.ModeRegisterDefinition = 0;

	/* Send the command */
	HAL_SDRAM_SendCommand(&sdramHandle, &Command, SDRAM_TIMEOUT);

	/* Step 4: Configure an Auto Refresh command */
	Command.CommandMode = FMC_SDRAM_CMD_AUTOREFRESH_MODE;
	Command.CommandTarget = FMC_SDRAM_CMD_TARGET_BANK1;
	Command.AutoRefreshNumber = 8;
	Command.ModeRegisterDefinition = 0;

	/* Send the command */
	HAL_SDRAM_SendCommand(&sdramHandle, &Command, SDRAM_TIMEOUT);

	/* Step 5: Program the external memory mode register */
	Command.CommandMode = FMC_SDRAM_CMD_LOAD_MODE;
	Command.CommandTarget = FMC_SDRAM_CMD_TARGET_BANK1;
	Command.AutoRefreshNumber = 1;
	Command.ModeRegisterDefinition = SDRAM_MODEREG_BURST_LENGTH_1 |
	    SDRAM_MODEREG_CAS_LATENCY_3;

	/* Send the command */
	HAL_SDRAM_SendCommand(&sdramHandle, &Command, SDRAM_TIMEOUT);

	/* Step 6: Program the Normal mode */
	Command.CommandMode = FMC_SDRAM_CMD_NORMAL_MODE;
	Command.CommandTarget = FMC_SDRAM_CMD_TARGET_BANK1;
	Command.AutoRefreshNumber = 1;
	Command.ModeRegisterDefinition = 0;

	/* Send the command */
	HAL_SDRAM_SendCommand(&sdramHandle, &Command, SDRAM_TIMEOUT);

	/* Step 7: Set the refresh rate counter */
	/* Set the device refresh rate */
	HAL_SDRAM_ProgramRefreshRate(&sdramHandle, RefreshCount);
}
